/*
 * WonderJS - A tiny game framework written in Javascript.
 *
 * copyright (c) 2012 Fabien LOISON <http://www.flogisoft.com/>
 * 
 * this program is free software. it comes without any warranty, to
 * the extent permitted by applicable law. you can redistribute it
 * and/or modify it under the terms of the do what the fuck you want
 * to public license, version 2, as published by sam hocevar. see
 * http://sam.zoy.org/wtfpl/copying for more details. 
 */


/**
 * A tiny game framework written in Javascript.
 *
 * @module WonderJS
 * @namespace wjs
 * @main WonderJS
 */


var wjs = {};


//////////////////////////////////////////////////////////////////////////////
// __main__
//////////////////////////////////////////////////////////////////////////////

/**
 * This class does not really exists. It justs group properties of the main
 * namspace of WonderJS (wjs).
 *
 * If you want to access to a property, you should use the
 * following syntax: `wjs.property` (e.g. `wjs.loader`).
 *
 * @class __main__
 */

/**
 * Instance of {{#crossLink "wjs.Loader"}}{{/crossLink}}. You
 * can use it for loading resources and scripts.
 *
 * @example
 *     wjs.loader.loadScript('js/game.js');
 *
 * @property loader
 * @type wjs.Loader
 */
wjs.loader = null;

/**
 * Instance of {{#crossLink "wjs.Helpers"}}{{/crossLink}}.
 *
 * @property helpers
 * @type wjs.Helpers
 */
wjs.loader = null;

/**
 * Instance of {{#crossLink "wjs.Loop"}}{{/crossLink}}.
 *
 * @property mainLoop
 * @type wjs.Loop
 */
wjs.mainLoop = null;

/**
 * Instance of {{#crossLink "wjs.KeyboardEvent"}}{{/crossLink}}.
 *
 * @property kbEvent
 * @type wjs.KeyboardEvent
 */
wjs.kbEvent = null;

// Some private properties
wjs._mainDiv = null;
wjs._canvasSize = null;
wjs._canvas = null;
wjs._ctx2d = null;

/**
 * Initialize WonderJS.
 *
 * @method init
 *
 * @param {String} divId The id of the `<div>` that will contain the `<canvas>`.
 * @param {Number} canvasWidth The width of the canvas (in pixel).
 * @param {Number} canvasHeight The height of the canvas (in pixel).
 */
wjs.init = function(divId, canvasWidth, canvasHeight)
{
	wjs._mainDiv = document.getElementById(divId);
	wjs._canvasSize = [canvasWidth, canvasHeight];
	wjs.loader = new wjs.Loader();
	wjs.helpers = new wjs.Helpers();
	wjs.kbEvent = new wjs.KeyboardEvent();
	wjs.mainLoop = new wjs.Loop();
}


//////////////////////////////////////////////////////////////////////////////
// Loader
//////////////////////////////////////////////////////////////////////////////

/**
 * Load scripts and resources.
 *
 * @class Loader
 * @constructor
 */
wjs.Loader = function()
{
	this._scripts = {};
	this._images = {};
}

/**
 * Load a Javascript file.
 *
 * @method loadScript
 *
 * @param {String} url The script url (e.g. `"./js/script.js"`).
 */
wjs.Loader.prototype.loadScript = function(url)
{
	this._scripts[url] = {'loaded': false};

	var e_script = document.createElement('script');
	e_script.src = url;
	e_script.type = 'text/javascript';
	e_script.onload = function() {
		this._scripts[url]['loaded'] = true;
	}.bind(this);

	document.getElementsByTagName('head')[0].appendChild(e_script);
}

/**
 * Load an image.
 *
 * @method loadImage
 *
 * @param {String} imageId An unique identifier for the image.
 * @param {String} url The image url.
 */
wjs.Loader.prototype.loadImage = function(imageId, url)
{
	this._images[imageId] = {'loaded': false, 'res': new Image()};

	this._images[imageId]['res'].src = url;
	this._images[imageId]['res'].onload = function() {
		this._images[imageId]['loaded'] = true;
	}.bind(this);
}

/**
 * Get a loaded image.
 *
 * @method getImage
 *
 * @param {String} imageId The unique identifier of the image.
 *
 * @return {Image} The image.
 */
wjs.Loader.prototype.getImage = function(imageId)
{
	return this._images[imageId]['res'];
}

/**
 * Wait until everything is loaded, and call the callback function.
 *
 * @method waitLoad
 *
 * @param {Function} callback The callback function.
 */
wjs.Loader.prototype.waitLoad = function(callback)
{
	// Check javascripts
	for (script in this._scripts)
	{
		if (!this._scripts[script]['loaded'])
		{
			setTimeout(this.waitLoad.bind(this, callback), 20);
			return;
		}
	}

	// Check images
	for (imageId in this._images)
	{
		if (!this._images[imageId]['loaded'])
		{
			setTimeout(this.waitLoad.bind(this, callback), 20);
			return;
		}
	}

	// Create the <canvas> in the main <div>
	wjs.helpers.clearMainDiv();
	wjs.helpers.createCanvas();
	// Start the main loop
	wjs.mainLoop.start();
	// Let's go!
	callback();
}


//////////////////////////////////////////////////////////////////////////////
// Loop
//////////////////////////////////////////////////////////////////////////////

/**
 * This class provide a loop that should run at ~60Hz.
 *
 * @class Loop
 * @constructor
 */
wjs.Loop = function()
{
	this._stopped = true;
	this._callbacks = {};
}

/**
 * Start the loop.
 *
 * @method start
 */
wjs.Loop.prototype.start = function()
{
	this._stopped = false;
	this._loop();
}

/**
 * Stop the loop.
 *
 * @method stop
 */
wjs.Loop.prototype.stop = function()
{
	this._stopped = true;
}

/**
 * Register a new callback to the loop. It will be called at each iteration
 * of the loop.
 *
 * @method addCallback
 *
 * @param {String} callbackId An unique identifier for the callback.
 * @param {Function} callback The callback function.
 */
wjs.Loop.prototype.addCallback = function(callbackId, callback)
{
	this._callbacks[callbackId] = callback;
}

/**
 * Unregister a callback from the loop.
 *
 * @method delCallback
 *
 * @param {String} callbackId The unique identifier of the callback.
 */
wjs.Loop.prototype.delCallback = function(callbackId)
{
	delete this._callbacks[callbackId];
}

/**
 * The loop.
 *
 * @method _loop
 * @private
 */
wjs.Loop.prototype._loop = function()
{
	if (!this._stopped)
	{
		// Call all the callbacks
		for (var callbackId in this._callbacks)
		{
			this._callbacks[callbackId]();
		}
		// Schedule next iteration
		if (window.requestAnimationFrame)
		{
			window.requestAnimationFrame(this._loop.bind(this));
		}
		else if (window.webkitRequestAnimationFrame) // Webkit
		{
			window.webkitRequestAnimationFrame(this._loop.bind(this));
		}
		else if (window.mozRequestAnimationFrame) // Firefox
		{
			window.mozRequestAnimationFrame(this._loop.bind(this));
		}
		else if (window.msRequestAnimationFrame) // MSIE
		{
			window.msRequestAnimationFrame(this._loop.bind(this));
		}
		else  // Fallback
		{
			setTimeout(this._loop.bind(this), 17);
		}
	}
}


//////////////////////////////////////////////////////////////////////////////
// Tileset
//////////////////////////////////////////////////////////////////////////////

/**
 * Class that manages a single tileset.
 *
 * @class Tileset
 * @constructor
 *
 * @param {Image} image The tileset image.
 * @param {Object} tiles The informations about the tiles that compose the 
 *        tileset.
 * @example
 *     {
 *         'tileId' {
 *             'size': [width, height], // Size of the tile
 *             'pos': [[x, y], ...],    // Position of all the frames of the tile (if animated)
 *             'speed': 5,              // Speed of the animation (0..60 ; 0 = not animated, 60 = 1sec if refresh rate = 60Hz)
 *             'step': 0                // First step of the animation (probably 0)...
 *         },
 *         ...
 *     }
 */
wjs.Tileset = function(image, tiles)
{
	this._image = image;
	this._tiles = tiles;

	this._step = 0;
}

/**
 * Blit the desired tile of the tileset.
 *
 * @method blit
 *
 * @param {String} tileId The tile idnetifier in the tileset.
 * @param {Number} x The x position where the tile will be drawn on the canvas.
 * @param {Number} y The y position where the tile will be drawn on the canvas.
 */
wjs.Tileset.prototype.blit = function(tileId, x, y)
{
	var tile = this._tiles[tileId];
	var step = tile['step'];

	wjs._ctx2d.drawImage(
		this._image,           // Image
		tile['pos'][step][0],  // Source image x
		tile['pos'][step][1],  // Source image y
		tile['size'][0],       // Source width
		tile['size'][1],       // Source height
		x,                     // Destination canvas x
		y,                     // Destination canvas y
		tile['size'][0],       // Destination width
		tile['size'][1]);      // Destination height
}

/**
 * Update the state of animated tiles.
 *
 * @method update
 */
wjs.Tileset.prototype.update = function()
{
	for (var tileId in this._tiles)
	{
		if (this._tiles[tileId]['speed'] > 0 && this._tiles[tileId]['pos'].length > 1)
		{
			if (this._step % this._tiles[tileId]['speed'] == 0)
			{
				this._tiles[tileId]['step'] += 1;
				if (this._tiles[tileId]['step'] >= this._tiles[tileId]['pos'].length)
				{
					this._tiles[tileId]['step'] = 0;
				}
			}
		}
	}

	this._step += 1;
	if (this._step >= 60)
	{
		this._step = 0;
	}
}


//////////////////////////////////////////////////////////////////////////////
// Map
//////////////////////////////////////////////////////////////////////////////

/**
 * A {{#crossLink "wjs.Map"}}{{/crossLink}} is a kind of grid.
 *
 * That permit you to make patterns with tiles from a same
 * {{#crossLink "wjs.Tileset"}}{{/crossLink}} and blit them on a canvas.
 *
 * @class Map
 * @constructor
 *
 * @param {wjs.Tileset} tileset The tileset.
 * @param {Number} cellWidth The width of each cell of the "grid".
 * @param {Number} cellHeight The height of each cell of the "grid".
 * @param {Array} map A two bidimensional array that contain in each "cell"
 *        the `tileId` of the tile that will blitted à this position. A `null`
 *        value means that the cell is empty.
 *
 * @example
 *     // A 3x3 map:
 *     [
 *         ['tileId', 'tileId', 'tileId'],
 *         ['tileId', null,     'tileId'],
 *         ['tileId', 'tileId', 'tileId']
 *     ]
 */
wjs.Map = function(tileset, cellWidth, cellHeight, map)
{
	this._tileset = tileset;
	this._cellWidth = cellWidth;
	this._cellHeight = cellHeight;
	this._map = map;
}

/**
 * Draw the all the tiles that compose the map on the canvas.
 *
 * @method draw
 */
wjs.Map.prototype.draw = function()
{
	for (var y in this._map)
	{
		for (var x in this._map[y])
		{
			if (this._map[y][x])
			{
				this._tileset.blit(
						this._map[y][x],
						x * this._cellWidth,
						y * this._cellHeight);
			}
		}
	}
}


//////////////////////////////////////////////////////////////////////////////
// KeyboardEvents
//////////////////////////////////////////////////////////////////////////////


/**
 * Handle keyboard events.
 *
 * @class KeyboardEvent
 * @constructor
 */
wjs.KeyboardEvent = function()
{
	this._keyDownCallbacks = {};
	this._keyUpCallbacks = {};

	// Bind the events
	window.addEventListener('keydown', this._onKeyDown.bind(this), false);
	window.addEventListener('keyup', this._onKeyUp.bind(this), false);
}

/**
 * Register a new callback for the `keydown` event.
 *
 * @method addKeyDownCallback
 *
 * @param {String} callbackId An unique identifier for the callback.
 * @param {Function} callback The callback function (the first parameter is
 *        the `keyName`).
 *
 * @example
 *     // Example of callback function
 *     function myCallback(keyName)
 *     {
 *         alert(keyName);
 *     }
 */
wjs.KeyboardEvent.prototype.addKeyDownCallback = function(callbackId, callback)
{
	this._keyDownCallbacks[callbackId] = callback;
}

/**
 * Unregister a callback from the `keydown` event.
 *
 * @method delKeyDownCallback
 *
 * @param {String} callbackId The unique identifier of the callback.
 */
wjs.KeyboardEvent.prototype.delKeyDownCallback = function(callbackId)
{
	delete this._keyDownCallbacks[callbackId];
}

/**
 * Register a new callback for the `keyup` event.
 *
 * @method addKeyUpCallback
 *
 * @param {String} callbackId An unique identifier for the callback.
 * @param {Function} callback The callback function (the first parameter is
 *        the `keyName`).
 *
 * @example
 *     // Example of callback function
 *     function myCallback(keyName)
 *     {
 *         alert(keyName);
 *     }
 */
wjs.KeyboardEvent.prototype.addKeyUpCallback = function(callbackId, callback)
{
	this._keyUpCallbacks[callbackId] = callback;
}

/**
 * Unregister a callback from the `keyup` event.
 *
 * @method delKeyUpCallback
 *
 * @param {String} callbackId The unique identifier of the callback.
 */
wjs.KeyboardEvent.prototype.delKeyUpCallback = function(callbackId)
{
	delete this._keyUpCallbacks[callbackId];
}

/**
 * Handle the `keydown` event and call the callback functions.
 *
 * @method _onKeyDown
 * @private
 *
 * @param {Event} event The event.
 */
wjs.KeyboardEvent.prototype._onKeyDown = function(event)
{
	var keyName = this._translateKey(event.keyCode);
	for (var callbackId in this._keyDownCallbacks)
	{
		this._keyDownCallbacks[callbackId](keyName);
	}
}

/**
 * Handle the `keyup` event and call the callback functions.
 *
 * @method _onKeyUp
 * @private
 *
 * @param {Event} event The event.
 */
wjs.KeyboardEvent.prototype._onKeyUp = function(event)
{
	var keyName = this._translateKey(event.keyCode);
	for (var callbackId in this._keyUpCallbacks)
	{
		this._keyUpCallbacks[callbackId](keyName);
	}
}

/**
 * Translate a keycode into a more understandable name.
 *
 * @method _translateKey
 * @private
 *
 * @param {Number} keycode The keycode.
 *
 * @return {String} the key name.
 */
wjs.KeyboardEvent.prototype._translateKey = function(keycode)
{
	switch (keycode)
	{
		case 37:
			return 'left';
			break;
		case 38:
			return 'up';
			break;
		case 39:
			return 'right';
			break;
		case 40:
			return 'down';
			break;
		default:
			return String.fromCharCode(keycode);
			break;
	}
}



//////////////////////////////////////////////////////////////////////////////
// Helpers
//////////////////////////////////////////////////////////////////////////////

/**
 * Load scripts and resources.
 *
 * @class Helpers
 * @constructor
 */
wjs.Helpers = function()
{
	// pass
}

/**
 * Clear the content of the main `<div>`.
 *
 * @method clearMainDiv
 */
wjs.Helpers.prototype.clearMainDiv = function()
{
	for (var i=0 ; i<wjs._mainDiv.childNodes.length ; i++)
	{
		wjs._mainDiv.removeChild(wjs._mainDiv.childNodes[i]);
	}
}

/**
 * Create the canvas inside the main `<div>`.
 *
 * @method createCanvas
 */
wjs.Helpers.prototype.createCanvas = function()
{
	wjs._canvas = document.createElement('canvas');
	wjs._canvas.width = wjs._canvasSize[0];
	wjs._canvas.height = wjs._canvasSize[1];
	wjs._mainDiv.appendChild(wjs._canvas);
	wjs._ctx2d = wjs._canvas.getContext('2d');
}
